import logging
import time
import math
import sys

import click
import click_log
import tqdm

import multiprocessing as mp
import concurrent.futures

import pysam

import longbow.utils.constants
from ..utils import model as LongbowModel
from ..utils.model import LibraryModel

from ..utils import bam_utils

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger("train")
click_log.basic_config(logger)


@click.command(name="train")
@click_log.simple_verbosity_option(logger)
@click.option(
    "-n",
    "--num-training-samples",
    type=int,
    default=10,
    show_default=True,
    help="number of training samples to use",
)
@click.option(
    "-i",
    "--max-training-iterations",
    type=int,
    default=5,
    show_default=True,
    help="number of training iterations to use",
)
@click.option(
    "-t",
    "--threads",
    type=int,
    default=mp.cpu_count() - 1,
    show_default=True,
    help="number of threads to use (0 for all)",
)
@click.option(
    "-o",
    "--output-yaml",
    required=True,
    type=click.Path(exists=False),
    help="trained model",
)
@click.option(
    "-m",
    "--model",
    default=longbow.utils.constants.DEFAULT_MODEL,
    show_default=True,
    help="The model to use for annotation.  If the given value is a pre-configured model name, then that "
         "model will be used.  Otherwise, the given value will be treated as a file name and Longbow will attempt to "
         "read in the file and create a LibraryModel from it.  Longbow will assume the contents are the configuration "
         "of a LibraryModel as per LibraryModel.to_json()."
)
@click.argument("training-bam", type=click.Path(exists=True))
def main(
    num_training_samples, max_training_iterations, threads, output_yaml, model, training_bam
):
    """Train transition and emission probabilities on real data."""

    t_start = time.time()

    logger.info("Invoked via: longbow %s", " ".join(sys.argv[1:]))

    threads = mp.cpu_count() if threads <= 0 or threads > mp.cpu_count() else threads
    logger.info(f"Running with {threads} worker subprocess(es)")

    # Get our model:
    if LibraryModel.has_prebuilt_model(model):
        m = LibraryModel.build_pre_configured_model(model)
    else:
        logger.info(f"Loading model from json file: %s", model)
        m = LibraryModel.from_json_file(model)
    logger.info(f"Using %s: %s", model, m.description)

    training_seqs = load_training_seqs(m, num_training_samples, threads, training_bam)

    logger.info("Loaded %d training sequences", len(training_seqs))

    logger.info("Starting training...", len(training_seqs))
    improvement, history = m.fit(
        sequences=training_seqs,
        max_iterations=max_training_iterations,
        stop_threshold=1e-1,
        return_history=True,
        verbose=True,
        n_jobs=threads,
    )

    with open(output_yaml, "w") as model_file:
        print(improvement.to_yaml(), file=model_file)

    logger.info(f"Done. Elapsed time: %2.2fs.", time.time() - t_start)


def load_training_seqs(m, num_training_samples, threads, training_bam):
    training_seqs = []
    raw_reads = []
    pysam.set_verbosity(0)  # silence message about the .bai file not being found
    with pysam.AlignmentFile(
        training_bam, "rb", check_sq=False, require_index=False
    ) as bam_file:
        for r in bam_file:
            raw_reads.append(r)

            if len(raw_reads) > num_training_samples:
                break
    with concurrent.futures.ThreadPoolExecutor(
        max_workers=threads
    ) as executor, tqdm.tqdm(
        desc="Progress", unit=" reads", colour="green", file=sys.stdout
    ) as pbar:

        future_to_segmented_read = {
            executor.submit(select_read, r, m): r for r in raw_reads
        }

        for future in concurrent.futures.as_completed(future_to_segmented_read):
            read = future_to_segmented_read[future]
            try:
                logp, seq = future.result()
                training_seqs.append(list(seq))
            except Exception as ex:
                logger.error("%r generated an exception: %s", read, ex)

            pbar.update(1)
    return training_seqs


def select_read(read, model):
    flogp = -math.inf
    fseq = None

    # Use the untrained model to determine if we should add this training
    # example in the forward or reverse-complement orientation.
    for seq in [read.query_sequence, bam_utils.reverse_complement(read.query_sequence)]:
        logp, ppath = model.annotate(seq, smooth_islands=True)

        if logp > flogp:
            flogp = logp
            fseq = seq

    return flogp, fseq



import gzip
# import pickle
# import numpy as np
# from sklearn.neighbors import NearestNeighbors

from ordered_set import OrderedSet

# import bz2
# import pickle 
# import _pickle as cPickle


def load_barcode_allowlist(filename):
    barcodes = OrderedSet()

    if filename is not None:
        f = gzip.open(filename, 'rb') if filename.endswith(".gz") else open(filename, 'rb')
        for l in f:
            bc = l.decode("utf-8").strip()

            barcodes.add(bc)

    return barcodes


import json
import math
import gzip
import logging
import click_log
import re
import pysam
import sys
import array
import collections
import os
import operator

from functools import reduce
from collections import OrderedDict
from construct import *
from inspect import currentframe, getdoc

from ..meta import VERSION

from .constants import RANDOM_SEGMENT_NAME, HPR_SEGMENT_TYPE_NAME, SEGMENTS_TAG, SEGMENTS_QUAL_TAG, SEGMENTS_RC_TAG, \
    SEGMENT_TAG_DELIMITER, READ_MODEL_NAME_TAG, READ_MODEL_SCORE_TAG, READ_APPROX_QUAL_TAG, READ_RAW_UMI_TAG, \
    READ_RAW_BARCODE_TAG, CONF_FACTOR_SCALE

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger("bam_utils")
click_log.basic_config(logger)

PB_READ_NAME_RE = re.compile("m[0-9]+e?_[0-9]{6}_[0-9]{6}/[0-9]+/.*")


# Named tuple to store alignment information:
class SegmentInfo(collections.namedtuple("SegmentInfo", ["name", "start", "end"])):

    _tag_regex = re.compile(r"(.*?):(\d+)-(\d+)")

    def __len__(self):
        # Must add 1 because the positions are inclusive coordinates:
        return self.end - self.start + 1

    def __str__(self):
        return f"SegmentInfo({self.to_tag()})"

    def to_tag(self):
        return f"{self.name}:{self.start}-{self.end}"

    @classmethod
    def from_tag(cls, tag_string):
        match = cls._tag_regex.match(tag_string)
        return SegmentInfo(match[1], int(match[2]), int(match[3]))


def load_read_count(pbi_file):
    """Compute file offsets for specified read names"""

    # Decode PacBio .pbi file.  This is not a full decode of the index, only the parts we need
    # until we get to the read count.
    # More on index format at https://pacbiofileformats.readthedocs.io/en/9.0/PacBioBamIndex.html .

    fmt = Struct(
        # Header
        "magic" / Const(b"PBI\x01"),
        "version_patch" / Int8ul,
        "version_minor" / Int8ul,
        "version_major" / Int8ul,
        "version_empty" / Int8ul,
        "pbi_flags" / Int16ul,
        "n_reads" / Int32ul,
        )

    with gzip.open(pbi_file, "rb") as f:
        idx_contents = fmt.parse_stream(f)

        return idx_contents.n_reads


def compute_shard_offsets(pbi_file, num_shards):
    """
    Compute all possible shard offsets (keeping adjacent reads from the ZMW together)
    """

    # Decode PacBio .pbi file.  This is not a full decode of the index, only the parts we need for sharding.
    # More on index format at https://pacbiofileformats.readthedocs.io/en/9.0/PacBioBamIndex.html .

    fmt = Struct(
        # Header
        "magic" / Const(b"PBI\x01"),
        "version_patch" / Int8ul,
        "version_minor" / Int8ul,
        "version_major" / Int8ul,
        "version_empty" / Int8ul,
        "pbi_flags" / Int16ul,
        "n_reads" / Int32ul,
        "reserved" / Padding(18),

        # Basic information section (columnar format)
        "rgId" / Padding(this.n_reads * 4),
        "qStart" / Padding(this.n_reads * 4),
        "qEnd" / Padding(this.n_reads * 4),
        "holeNumber" / Array(this.n_reads, Int32sl),
        "readQual" / Padding(this.n_reads * 4),
        "ctxtFlag" / Padding(this.n_reads * 1),
        "fileOffset" / Array(this.n_reads, Int64sl),
        )

    # Make a list of bgzf virtual file offsets for sharding and store ZMW counts.
    file_offsets_hash = OrderedDict()
    last_offset = 0
    zmw_count_hash = dict()
    with gzip.open(pbi_file, "rb") as f:
        idx_contents = fmt.parse_stream(f)

        for j in range(0, idx_contents.n_reads):
            # Save only the virtual file offset for the first ZMW hole number, so
            # that shard boundaries always keep reads from the same ZMW together.
            if idx_contents.holeNumber[j] not in file_offsets_hash:
                file_offsets_hash[idx_contents.holeNumber[j]] = idx_contents.fileOffset[j]

            last_offset = idx_contents.fileOffset[j]

            try:
                zmw_count_hash[idx_contents.holeNumber[j]] += 1
            except KeyError:
                zmw_count_hash[idx_contents.holeNumber[j]] = 1

    file_offsets = list(file_offsets_hash.values())
    shard_offsets = []
    read_counts = []
    read_nums = []

    read_num = 1
    by = int(math.ceil(len(file_offsets) / num_shards))
    for i in range(0, len(file_offsets), by):
        shard_offsets.append(file_offsets[i])
        read_counts.append(len(file_offsets[i:i + by]))
        read_nums.append(read_num)
        read_num += read_counts[-1]

    # For the last read in the file, pad the offset so the final comparison in write_shard() retains the final read.
    shard_offsets.append(file_offsets[-1] + 1)

    return shard_offsets, zmw_count_hash, idx_contents.n_reads, read_counts, read_nums


def create_bam_header_with_program_group(command_name, base_bam_header, description=None, models=None):
    """Create a pysam.AlignmentHeader object with program group (PG) information populated by the given arguments.

    This function is intended to be called from the 'main' function of a longbow subcommand because it uses reflection
    to pull in the first line of the docstring from the main function as the description (DS field)."""

    bam_header_dict = base_bam_header.to_dict()

    if not description:
        prev_frame = currentframe().f_back
        description = getdoc(prev_frame.f_globals['main']).split("\n")[0]

    # If we have a model here, we should add the description of the model to our program group:
    if models:
        description = description + "  MODEL(s): " + ", ".join([m.to_json(indent=None) for m in models])

    # Add our program group to it:
    pg_dict = {
        "ID": f"longbow-{command_name}-{VERSION}",
        "PN": "longbow",
        "VN": f"{VERSION}",
        # Use reflection to get the first line of the doc string the caller - the main function for our header:
        "DS": description,
        "CL": " ".join(sys.argv),
    }
    if "PG" in bam_header_dict:
        bam_header_dict["PG"].append(pg_dict)
    else:
        bam_header_dict["PG"] = [pg_dict]
    out_header = pysam.AlignmentHeader.from_dict(bam_header_dict)

    return out_header


def check_for_preexisting_files(file_list, exist_ok=False):
    """Checks if the files in the given file_list exist.
    If any file exists and exist_ok is False, this will exit the program.
    If any file exists and exist_ok is True, the program will continue.
    """

    # Allow users to be a little lazy with what input types they give:
    if not isinstance(file_list, list) and not isinstance(file_list, set):
        file_list = [file_list]

    do_files_exist = False
    for f in file_list:
        if os.path.exists(f) and not f == "/dev/null":
            if exist_ok:
                logger.warning(f"Output file exists: {f}.  Overwriting.")
            else:
                logger.error(f"Output file already exists: {f}!")
                do_files_exist = True
    if do_files_exist:
        sys.exit(1)


def get_segment_score(read_sequence, segment, library_model, ssw_aligner=None):
    """Get the alignment score of the given segment against the read sequence."""

    # TODO: FIX THIS METHOD WITH NEW SCORING MODEL FROM INSPECT!
    # It should be something like:
    #                         base, count = library_model.adapter_dict[lbl][special_seg_type]
    #                         known_segment_seq = base * count
    #                         segment_bases = _get_segment_bases(seq, total_bases_seen, segments)
    #
    #                         seg_score_string = f" ({len(known_segment_seq) - editdistance.eval(segment_bases, known_segment_seq)}" \
    #                                            f"/{len(known_segment_seq)})"


    # We don't score random segments:
    if segment.name == RANDOM_SEGMENT_NAME or \
            (library_model.has_named_random_segments and segment.name in library_model.named_random_segments):
        return 0, 0

    # Create a default aligner if we weren't given one:
    if not ssw_aligner:
        ssw_aligner = ssw.Aligner()

    # get our model segment so we can reference it later:
    if type(library_model.adapter_dict[segment.name]) == str:
        model_seg_sequence = library_model.adapter_dict[segment.name]
    elif type(library_model.adapter_dict[segment.name]) == dict:
        if list(library_model.adapter_dict[segment.name].keys())[0] == HPR_SEGMENT_TYPE_NAME:
            b, l = list(library_model.adapter_dict[segment.name].values())[0]
            model_seg_sequence = b * l
        else:
            raise RuntimeError(f"Cannot calculate score for dict segment type: "
                               f"{type(library_model.adapter_dict[segment.name])}: {segment}")
    else:
        raise RuntimeError(
            f"Unknown segment type: {type(library_model.adapter_dict[segment.name])} for segment: {segment}"
        )

    # Get our alignment and our score:
    if segment.end - segment.start > 1:
        alignment = ssw_aligner.align(read_sequence[segment.start:segment.end], model_seg_sequence)
        optimal_score = alignment.score
    else:
        optimal_score = 0

    # The max score is the match score * the length of the reference segment
    max_score = len(model_seg_sequence) * ssw_aligner.matrix.get_match()

    return optimal_score, max_score


def collapse_annotations(path):
    """Collapses given path into a list of SegmentInfo objects."""
    last = ""
    start = 0
    segments = []
    i = 0
    for i, seg in enumerate(path):
        if seg != last:
            if i != 0:
                segments.append(SegmentInfo(last, start, i - 1))
            last = seg
            start = i
    # Don't forget the last one:
    segments.append(SegmentInfo(last, start, i))

    return segments


def write_annotated_read(read, segments, is_rc, logp, model, ssw_aligner, out_bam_file):
    """Write the given pysam.AlignedSegment read object to the given file with the given metadata."""

    # Obligatory log message:
    logger.debug(
        "Path for read %s (%2.2f)%s: %s",
        read.query_name,
        logp,
        " (RC)" if is_rc else "",
        segments,
    )

    # Set our tag and write out the read to the annotated file:
    read.set_tag(SEGMENTS_TAG, SEGMENT_TAG_DELIMITER.join([s.to_tag() for s in segments]))

    # Set the model info tags:
    read.set_tag(READ_MODEL_SCORE_TAG, logp)
    read.set_tag(READ_MODEL_NAME_TAG, model.name)

    # If we're reverse complemented, we make it easy and just reverse complement the read and add a tag saying
    # that the read was RC:
    read.set_tag(SEGMENTS_RC_TAG, is_rc)
    if is_rc:
        quals = read.query_qualities[::-1]
        seq = reverse_complement(read.query_sequence)
        read.query_sequence = seq
        read.query_qualities = quals

    # Get our segment scores and set them:
    total_score = 0
    total_max_score = 0
    score_strings = []
    for s in segments:
        score, max_score = get_segment_score(read.query_sequence, s, model, ssw_aligner)
        score_strings.append(f"{score}/{max_score}")
        total_score += score
        total_max_score += max_score

    read.set_tag(SEGMENTS_QUAL_TAG, SEGMENT_TAG_DELIMITER.join(score_strings))
    if total_max_score != 0:
        read.set_tag(READ_APPROX_QUAL_TAG, f"{total_score / total_max_score:.4f}")
    else:
        read.set_tag(READ_APPROX_QUAL_TAG, f"0.0")

    out_bam_file.write(read)


# IUPAC RC's from: http://arep.med.harvard.edu/labgc/adnan/projects/Utilities/revcomp.html
# and https://www.dnabaser.com/articles/IUPAC%20ambiguity%20codes.html
RC_BASE_MAP = {
    "N": "N",
    "A": "T",
    "T": "A",
    "G": "C",
    "C": "G",
    "Y": "R",
    "R": "Y",
    "S": "S",
    "W": "W",
    "K": "M",
    "M": "K",
    "B": "V",
    "V": "B",
    "D": "H",
    "H": "D",
    "n": "n",
    "a": "t",
    "t": "a",
    "g": "c",
    "c": "g",
    "y": "r",
    "r": "y",
    "s": "s",
    "w": "w",
    "k": "m",
    "m": "k",
    "b": "v",
    "v": "b",
    "d": "h",
    "h": "d",
}


def reverse_complement(base_string):
    """
    Reverse complements the given base_string.
    :param base_string: String of bases to be reverse-complemented.
    :return: The reverse complement of the given base string.
    """

    return "".join(map(lambda b: RC_BASE_MAP[b], base_string[::-1]))


def get_confidence_factor(qual_string: str, scale_factor: float = CONF_FACTOR_SCALE) -> float:
    """Get the confidence factor for the given sequence to be tallied for use with STARCODE.
    quals are assumed to be phred scale quality scores in string format and will be converted to numerical values."""
    return scale_factor * reduce(
        operator.mul, map(lambda q: 1. - 10 ** (-(ord(q) - 33.) / 10), qual_string)
    )


def get_confidence_factor_raw_quals(quals: array.array, scale_factor: float = CONF_FACTOR_SCALE) -> float:
    """Get the confidence factor for the given sequence to be tallied for use with STARCODE.
    quals are assumed to be numerical already and will not be type converted."""
    return scale_factor * reduce(
        operator.mul, map(lambda q: 1. - 10 ** (-q/10), quals)
    )


def has_cbc_and_umi(read):
    return read.has_tag(READ_RAW_BARCODE_TAG) and read.has_tag(READ_RAW_UMI_TAG)


def get_model_name_from_bam_header(header):
    return get_model_name_from_bam_header(header)['name']


def get_model_from_bam_header(header):
    for pg in header.as_dict()['PG']:
        if pg['PN'] == 'longbow' and 'annotate' in pg['ID']:
            desc, models_str= pg['DS'].split('MODEL(s): ')
            models_json = json.loads(models_str)
            return models_json

    return None


from cmath import inf
import logging
import math
import sys
import itertools
import re
import time
import os

import click
import click_log
import tqdm

import numpy as np
import pysam
import multiprocessing as mp

import gzip
from construct import *

import longbow.utils.constants
from ..utils import bam_utils
from ..utils.bam_utils import SegmentInfo
from ..utils import model as LongbowModel
from ..utils.model import LibraryModel


logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger("peek")
click_log.basic_config(logger)


@click.command(name=logger.name)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-p",
    "--pbi",
    required=False,
    type=click.Path(),
    help="BAM .pbi index file",
)
@click.option(
    "-t",
    "--threads",
    type=int,
    default=mp.cpu_count() - 1,
    show_default=True,
    help="number of threads to use (0 for all)",
)
@click.option(
    "-o",
    "--output-model",
    default="-",
    type=click.Path(exists=False),
    help="model name output  [default: stdout]",
)
@click.option(
    "-c",
    "--chunk",
    type=str,
    default="",
    required=False,
    help="Process a single chunk of data (e.g. specify '2/4' to process the second of four equally-sized "
         "chunks across the dataset)"
)
@click.option(
    "-n",
    "--num-reads",
    type=int,
    default=100,
    required=False,
    help="Number of reads to examine when trying to guess the array model"
)
@click.option(
    "-l",
    "--min-length",
    type=int,
    default=0,
    show_default=True,
    required=False,
    help="Minimum length of a read to process.  Reads shorter than this length will not be annotated."
)
@click.option(
    "-L",
    "--max-length",
    type=int,
    default=longbow.utils.constants.DEFAULT_MAX_READ_LENGTH,
    show_default=True,
    required=False,
    help="Maximum length of a read to process.  Reads longer than this length will not be annotated."
)
@click.option(
    "--min-rq",
    type=float,
    default=-2,
    show_default=True,
    required=False,
    help="Minimum ccs-determined read quality for a read to be annotated.  CCS read quality range is [-1,1]."
)
@click.option(
    '-f',
    '--force',
    is_flag=True,
    default=False,
    show_default=True,
    help="Force overwrite of the output files if they exist."
)
@click.option(
    "-d",
    "--include-deprecated-models",
    is_flag=True,
    default=False,
    show_default=True,
    help="Examine the deprecated built-in models as well",
)
@click.argument("input-bam", default="-" if not sys.stdin.isatty() else None, type=click.File("rb"))
def main(pbi, threads, output_model, chunk, num_reads, min_length, max_length, min_rq, force, include_deprecated_models, input_bam):
    """Peeks at MAS-ISO-seq reads to guess the most appropriate pre-built array model to use."""

    t_start = time.time()

    logger.info("Invoked via: longbow %s", " ".join(sys.argv[1:]))

    # Check to see if the output files exist:
    bam_utils.check_for_preexisting_files(output_model, exist_ok=force)

    threads = mp.cpu_count() if threads <= 0 or threads > mp.cpu_count() else threads
    logger.info(f"Running with {threads} worker subprocess(es)")

    # Make all prebuilt models
    models = {}
    for model_name in LibraryModel.pre_configured_models:
        if not LibraryModel.pre_configured_models[model_name]['deprecated'] or include_deprecated_models:
            m = LibraryModel.build_pre_configured_model(model_name)
            models[model_name] = m

    pbi = f"{input_bam.name}.pbi" if pbi is None else pbi
    read_count = None
    read_num = 0
    start_offset = 0
    end_offset = math.inf

    if not os.path.exists(pbi) and chunk is not "":
        raise ValueError(f"Chunking specified but pbi file '{pbi}' not found")

    if os.path.exists(pbi):
        if chunk is not "":
            (chunk, num_chunks) = re.split("/", chunk)
            chunk = int(chunk)
            num_chunks = int(num_chunks)

            # Decode PacBio .pbi file and determine the shard offsets.
            offsets, zmw_counts, read_count, read_counts_per_chunk, read_nums = bam_utils.compute_shard_offsets(pbi, num_chunks)

            start_offset = offsets[chunk - 1]
            end_offset = offsets[chunk] if chunk < len(offsets) else offsets[chunk - 1]
            read_count = read_counts_per_chunk[chunk - 1] if chunk < len(offsets) else 0
            read_num = read_nums[chunk - 1] if chunk < len(offsets) else 0

            logger.info("Detecting best model in %d reads from chunk %d/%d (reads %d-%d)", min(read_count, num_reads), chunk, num_chunks, read_num, read_num + read_count - 1)
        else:
            read_count = bam_utils.load_read_count(pbi)
            logger.info("Detecting best model in %d reads", min(read_count, num_reads))
    else:
        logger.info("Detecting best model in %d reads", num_reads)

    # Create queues for data:
    queue_size = threads * 2 if threads < 10 else 20
    manager = mp.Manager()
    input_data_queue = manager.Queue(maxsize=queue_size)
    results = manager.Queue()

    # Start worker sub-processes:
    worker_pool = []

    for i in range(threads):
        p = mp.Process(
            target=_worker_attempt_segmentations_fn, args=(input_data_queue, results, i, min_length, max_length, min_rq, models)
        )
        p.start()
        worker_pool.append(p)

    pysam.set_verbosity(0)  # silence message about the .bai file not being found
    with pysam.AlignmentFile(
        input_bam if start_offset == 0 else input_bam.name, "rb", check_sq=False, require_index=False
    ) as bam_file:

        # If we're chunking, advance to the specified virtual file offset.
        if start_offset > 0:
            bam_file.seek(start_offset)

        # Start output worker:
        res = manager.dict()
        output_worker = mp.Process(
            target=_collect_thread_fn,
            args=(results, output_model, not sys.stdin.isatty(), res, num_reads if read_count is None else min(read_count, num_reads))
        )
        output_worker.start()

        # Add in a sentinel value at the end of the queue - one for each subprocess - so we guarantee
        # that all subprocesses will exit:
        reads_seen = 0
        iter_data = itertools.chain(bam_file, (None,) * threads)
        for r in iter_data:
            # We have to adjust for our sentinel value if we've got to it:
            if r is not None:
                r = r.to_string()
            input_data_queue.put(r)

            if start_offset > 0:
                if bam_file.tell() >= end_offset or r is None:
                    [input_data_queue.put(None) for _ in range(threads)]
                    break

            if reads_seen >= num_reads:
                [input_data_queue.put(None) for _ in range(threads)]
                break

            # Increment reads seen
            reads_seen += 1

        logger.debug("Finished reading data and sending it to sub-processes.")
        logger.debug("Waiting for sub-processes to finish...")

        # Wait for our input jobs to finish:
        for p in worker_pool:
            p.join()

        logger.debug("All workers stopped.")
        logger.debug("Terminating output process.")

        # Now that our input processes are done, we can add our exit sentinel onto the output queue and
        # wait for that process to end:
        results.put(None)
        output_worker.join()

    logger.info("Histogram of most likely model counts:")

    res = dict(sorted(res.items(), key=lambda item: item[1], reverse=True))
    plot_model_counts(res, models)
    best_model, best_model_count = next(iter(res.items()))

    logger.info(f"Overall most likely model: {best_model} (seen in {best_model_count} reads, {100.0*best_model_count/np.sum(list(res.values())):.1f}%)")

    with open(output_model if output_model != "-" else "/dev/stdout", "w") as wm:
        wm.write(f'{best_model}\n')

    et = time.time()
    logger.info(f"Done. Elapsed time: {et - t_start:2.2f}s. "
                f"Overall processing rate: {reads_seen/(et - t_start):2.2f} reads/s.")


def plot_model_counts(res, models, max_width=50.0):
    big_block_char = u"\u2588"
    small_block_char = u"\u258F"

    model_tot = np.sum(list(res.values()))
    max_label_width = np.max(list(map(lambda x: len(x), models.keys())))
    for model_name in res:
        pct = 100.0 * res[model_name] / model_tot
        num_blocks = int(res[model_name] * max_width / model_tot)

        logger.info(f"  {model_name:>{max_label_width}} {big_block_char*(num_blocks+1)} {res[model_name]} ({pct:.1f}%)")

    for model_name in models:
        if model_name not in res:
            logger.info(f"  {model_name:>{max_label_width}} {small_block_char} 0 (0.0%)")


def get_segments(read):
    """Get the segments corresponding to a particular read by reading the segments tag information."""
    return read.to_string(), [
        SegmentInfo.from_tag(s) for s in read.get_tag(longbow.utils.constants.SEGMENTS_TAG).split(
            longbow.utils.constants.SEGMENT_TAG_DELIMITER)
    ]


def _collect_thread_fn(out_queue, out_bam_file_name, disable_pbar, res, read_count):
    """Thread / process fn to write out all our data."""

    with tqdm.tqdm(
            desc="Progress",
            unit=" read",
            colour="green",
            file=sys.stderr,
            disable=disable_pbar,
            total=read_count
        ) as pbar:

        while True:
            # Wait for some output data:
            raw_data = out_queue.get()

            # Check for exit sentinel:
            if raw_data is None:
                break
            # Should really never be None, but just in case:
            elif raw_data is None:
                continue

            # Unpack data:
            best_model = raw_data

            if best_model not in res:
                res[best_model] = 0
            res[best_model] += 1

            pbar.update(1)


def _worker_attempt_segmentations_fn(in_queue, out_queue, worker_num, min_length, max_length, min_rq, models):
    """Function to run in each subthread / subprocess.
    Segments each read and place the segments in the output queue."""

    num_reads_segmented = 0

    while True:
        # Wait until we get some data.
        # Note: Because we have a sentinel value None inserted at the end of the input data for each
        #       subprocess, we don't have to add a timeout - we're guaranteed each process will always have
        #       at least one element.
        raw_data = in_queue.get()

        # Check for exit sentinel:
        if raw_data is None:
            break
        # Should really never be None, but just in case:
        elif raw_data is None:
            continue

        # Unpack our data here:
        read = raw_data
        read = pysam.AlignedSegment.fromstring(
            read, pysam.AlignmentHeader.from_dict(dict())
        )

        # Check for min/max length and min quality:
        if len(read.query_sequence) < min_length:
            logger.debug(f"Read is shorter than min length.  "
                           f"Skipping: {read.query_name} ({len(read.query_sequence)} < {min_length})")
            continue
        elif len(read.query_sequence) > max_length:
            logger.debug(f"Read is longer than max length.  "
                           f"Skipping: {read.query_name} ({len(read.query_sequence)} > {max_length})")
            continue
        elif read.get_tag("rq") < min_rq:
            logger.debug(f"Read quality is below the minimum.  "
                           f"Skipping: {read.query_name} ({read.get_tag('rq')} < {min_rq})")
            continue

        # Process and place our data on the output queue:
        best_model = _get_best_model(read, models)

        out_queue.put(best_model)
        num_reads_segmented += 1

    logger.debug(f"Worker %d: Num reads segmented: %d", worker_num, num_reads_segmented)


def _get_best_model(read, models):
    model_scores = {}

    for model_name in models:
        model = models[model_name]
        fw_logp, _ = model.annotate(read.query_sequence)
        rc_logp, _ = model.annotate(bam_utils.reverse_complement(read.query_sequence))

        model_scores[f'{model_name}:fw'] = fw_logp
        model_scores[f'{model_name}:rc'] = rc_logp

    model_scores = dict(sorted(model_scores.items(), key=lambda item: item[1], reverse=True))

    best_model_full = next(iter(model_scores))
    best_model = re.sub(':.*$', '', best_model_full)

    return best_model



import logging
import time
import os
import sys

import click
import click_log

import tqdm
import pysam
from construct import *

import longbow.utils.constants
from ..utils import bam_utils
from ..utils import model as LongbowModel
from ..utils.model import LibraryModel

from ..utils.bam_utils import SegmentInfo

from ..annotate.command import get_segments
from ..segment.command import create_simple_delimiters
from ..segment.command import segment_read_with_simple_splitting
from ..segment.command import create_simple_split_array_element

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger("extract")
click_log.basic_config(logger)


@click.command(name=logger.name)
@click_log.simple_verbosity_option(logger)
@click.option(
    "-p",
    "--pbi",
    required=False,
    type=click.Path(exists=True),
    help="BAM .pbi index file",
)
@click.option(
    "-o",
    "--output-bam",
    default="-",
    type=click.Path(exists=False),
    help="extracted bam output.  [default: stdout]",
)
@click.option(
    '-f',
    '--force',
    is_flag=True,
    default=False,
    show_default=True,
    help="Force overwrite of the output files if they exist."
)
@click.option(
    "--create-barcode-conf-file",
    required=False,
    is_flag=True,
    default=False,
    help=f"Create a barcode confidence score file based on the barcodes in the given model.  "
         f"This only applies for models that have annotation_segments where one such segment "
         f"is annotated into the raw barcode field ({longbow.utils.constants.READ_RAW_BARCODE_TAG})",
)
@click.option(
    "-b",
    "--base-padding",
    default=2,
    required=False,
    show_default=True,
    type=int,
    help="Number of bases to include on either side of the extracted region(s).",
)
@click.option(
    "--leading-adapter",
    default=None,
    required=False,
    type=str,
    help="Adapter preceding the region to extract.  Required if the given model does not name a `coding_region`.",
)
@click.option(
    "--trailing-adapter",
    default=None,
    required=False,
    type=str,
    help="Adapter following the region to extract.  Required if the given model does not name a `coding_region`.",
)
@click.option(
    # CBC + UMI for MAS15 is 26
    "--start-offset",
    default=None,
    required=False,
    show_default=True,
    type=int,
    help="Number of bases to ignore from the extracted region start.  "
         "These bases will not be included in the extracted sequences.  "
         "Required if the given model does not name a `coding_region`.",
)
@click.option(
    "-m",
    "--model",
    help="The model to use for annotation.  If not specified, it will be autodetected from "
         "the BAM header.  If the given value is a pre-configured model name, then that "
         "model will be used.  Otherwise, the given value will be treated as a file name "
         "and Longbow will attempt to read in the file and create a LibraryModel from it.  "
         "Longbow will assume the contents are the configuration of a LibraryModel as per "
         "LibraryModel.to_json()."
)
@click.argument("input-bam", default="-" if not sys.stdin.isatty() else None, type=click.File("rb"))
def main(pbi, output_bam, force, base_padding, create_barcode_conf_file,
         leading_adapter, trailing_adapter, start_offset, model, input_bam):
    """Extract coding segments from the reads in the given bam.
    The main coding segments are assumed to be labeled as `random` segments or labeled with the same name as the
    `coding_region` in the model (if specified).

    For `random` segment extraction:
    Uses known segments flanking the region to be extracted as markers to indicate
    the start and end of what to extract.

    For `coding_region` segment extraction:
    Looks for any section of the reads labeled with the same name as the `coding_region` in the model, regardless of
    position.
    """

    t_start = time.time()

    logger.info("Invoked via: longbow %s", " ".join(sys.argv[1:]))

    # Check to see if the output file exists:
    bam_utils.check_for_preexisting_files(output_bam, exist_ok=force)

    logger.info(f"Writing extracted read segments to: {output_bam}")
    logger.info(f"Including {base_padding} flanking bases.")

    pbi = f"{input_bam.name}.pbi" if pbi is None else pbi
    read_count = None
    if os.path.exists(pbi):
        read_count = bam_utils.load_read_count(pbi)
        logger.info("About to Extract segments from %d reads", read_count)

    # Open our input bam file:
    pysam.set_verbosity(0)
    with pysam.AlignmentFile(input_bam, "rb", check_sq=False, require_index=False) as bam_file, \
            tqdm.tqdm(
            desc="Progress",
            unit=" read",
            colour="green",
            file=sys.stderr,
            disable=not sys.stdin.isatty(),
            total=read_count
            ) as pbar:

        # Get our model:
        if model is None:
            lb_model = LibraryModel.from_json_obj(bam_utils.get_model_from_bam_header(bam_file.header))
        elif model is not None and LibraryModel.has_prebuilt_model(model):
            lb_model = LibraryModel.build_pre_configured_model(model)
        else:
            lb_model = LibraryModel.from_json_file(model)

        logger.info(f"Using %s: %s", lb_model.name, lb_model.description)

        # Validate our command line arguments:
        if lb_model.has_coding_region:
            logger.info(f"Extracting coding region from model {lb_model.name}: {lb_model.coding_region}")
        else:
            # We don't have a model with a coding region.
            # We MUST make sure we have the other inputs that we need:
            missing_args = []

            if leading_adapter is None:
                missing_args.append("--leading-adapter")
            if trailing_adapter is None:
                missing_args.append("--trailing-adapter")
            if start_offset is None:
                missing_args.append("--start-offset")

            if len(missing_args) > 0:
                message = f"ERROR: the following arguments must be specified when using a model that has no " \
                        f"`coding_region`: {', '.join(missing_args)}"
                logger.critical(message)
                raise RuntimeError(message)

            logger.info(f"Extracting `{longbow.utils.constants.RANDOM_SEGMENT_NAME}` segments between {leading_adapter} and {trailing_adapter}.")
            logger.info(f"Ignoring the first {start_offset} bases from extracted read segments.")

        # Set up our barcode confidence file here:
        barcode_conf_file = None
        if create_barcode_conf_file:
            if lb_model.has_barcode_annotation:
                logger.info(f"Creating barcode confidence file: {longbow.utils.constants.BARCODE_CONF_FILE_NAME}")
                barcode_conf_file = open(longbow.utils.constants.BARCODE_CONF_FILE_NAME, 'w')
            else:
                logger.warning(f"Model does not have a barcode output, but barcode creation flag was given.  "
                               f"Barcode confidence file will NOT be created.")

        # Get our delimiters here just in case we have to split the reads later:
        delimiters = create_simple_delimiters(lb_model)

        # Get our header from the input bam file:
        out_header = bam_utils.create_bam_header_with_program_group(logger.name, bam_file.header)

        # Setup output files:
        with pysam.AlignmentFile(output_bam, "wb", header=out_header) as extracted_bam_file:

            num_reads = 0
            num_reads_with_extracted_segments = 0
            num_segments_extracted = 0
            num_segments_skipped = 0

            for read in bam_file:

                # Get our read segments:
                try:
                    _, segments = get_segments(read)
                except KeyError:
                    logger.error(f"Input bam file does not contain longbow segmented reads!  "
                                 f"No {longbow.utils.constants.SEGMENTS_TAG} tag detected on read {read.query_name} !")
                    sys.exit(1)

                # Check if the read is already segmented:
                if (not read.has_tag(longbow.utils.constants.READ_IS_SEGMENTED_TAG)) or (not read.get_tag(
                        longbow.utils.constants.READ_IS_SEGMENTED_TAG)):
                    # The read is not segmented.  We should segment it first and then go through our segments one by one:
                    logger.debug(f"Read must be segmented prior to extraction: {read.query_name}")
                    segmented_reads = []
                    segment_bounds_tuples = segment_read_with_simple_splitting(read, delimiters, segments)
                    for prev_delim_name, delim_name, start_coord, end_coord in segment_bounds_tuples:
                        segmented_reads.append(
                            create_simple_split_array_element(
                                delim_name, end_coord, lb_model, prev_delim_name, read, segments, start_coord
                            )
                        )
                else:
                    segmented_reads = [read]

                extracted_segment = False
                for r in segmented_reads:
                    segments = _get_segments_only(r)
                    # If our model has a coding region, we should use it:
                    if lb_model.has_coding_region:
                        es, nsx, nss = \
                            _extract_region_from_model_with_coding_region(r, segments, lb_model, base_padding, extracted_bam_file, barcode_conf_file)
                    else:
                        # We do not have a coding section, so we'll need to do things the old fashioned way:
                        es, nsx, nss = \
                            _extract_region_from_model_without_coding_region(r, segments, leading_adapter, trailing_adapter, start_offset, base_padding, extracted_bam_file, barcode_conf_file)

                    # Track our stats:
                    extracted_segment |= es
                    num_segments_extracted += nsx
                    num_segments_skipped += nss

                num_reads += 1
                if extracted_segment:
                    num_reads_with_extracted_segments += 1

                pbar.update(1)

    # Close our barcode file if it was opened:
    if barcode_conf_file is not None:
        barcode_conf_file.close()

    # Calc some stats:
    pct_reads_with_extracted_segments = 100 * num_reads_with_extracted_segments / num_reads if num_reads > 0 else 0
    segs_per_read = num_segments_extracted / num_reads if num_reads > 0 else 0

    # Yell at the user:
    logger.info(f"Done. Elapsed time: %2.2fs.", time.time() - t_start)
    logger.info(f"Total # Reads Processed: %d", num_reads)
    logger.info(f"# Reads Containing Extracted Segments: %d (%2.2f%%)",
                num_reads_with_extracted_segments, pct_reads_with_extracted_segments)
    logger.info(f"Total # Segments Extracted: %d", num_segments_extracted)
    logger.info(f"Total # Segments Skipped: %d", num_segments_skipped)
    logger.info(f"# Segments extracted per read: %2.2f", segs_per_read)


def _get_segments_only(read):
    """Get the segments corresponding to a particular read by reading the segments tag information."""
    return [
        SegmentInfo.from_tag(s) for s in read.get_tag(longbow.utils.constants.SEGMENTS_TAG).split(
            longbow.utils.constants.SEGMENT_TAG_DELIMITER)
    ]


def _extract_region_from_model_without_coding_region(array_element_read, segments, leading_adapter,
                                                     trailing_adapter, start_offset, base_padding, extracted_bam_file,
                                                     barcode_conf_file):
    # Get our marker segments:
    start_marker_list = [(i, s) for i, s in enumerate(segments) if s.name == leading_adapter]
    end_marker_list = [(i, s) for i, s in enumerate(segments) if s.name == trailing_adapter]

    if len(start_marker_list) != len(end_marker_list):
        logger.warning(f"Found %d start markers and %d end markers.  Only looking at first %d pairs.  "
                       f"(starts: %s, ends: %s)",
                       len(start_marker_list), len(end_marker_list),
                       min(len(start_marker_list), len(end_marker_list)),
                       " ".join([f"{i}:{s.name}" for i, s in start_marker_list]),
                       " ".join([f"{i}:{s.name}" for i, s in end_marker_list]))

    extracted_segment = False

    num_segments_skipped = 0
    num_segments_extracted = 0

    # Go through each marker pair and do the extraction:
    for s_info, e_info in zip(start_marker_list, end_marker_list):

        si, start_marker = s_info
        ei, end_marker = e_info

        # Does the start marker come before the end marker and do we have exactly one random segment in
        # between them?
        if (start_marker.end < end_marker.start) and (ei - si == 2) and (
                segments[si + 1].name == "random"):
            # We have a valid segment to extract:
            logger.debug("Found a segment to extract: %s: %s", array_element_read.query_name, segments[si + 1])

            # Create an AlignedSegment to output:
            aligned_segment = _create_extracted_aligned_segment(
                array_element_read, segments[si + 1], start_offset, base_padding, barcode_conf_file
            )

            if aligned_segment:
                extracted_bam_file.write(aligned_segment)
                num_segments_extracted += 1
                extracted_segment = True
            else:
                num_segments_skipped += 1
        else:
            if start_marker.end >= end_marker.start:
                logger.warning(
                    "Read %s: start marker segment (i=%d) occurs at or after end segment (i=%d):"
                    " %d >= %d.  Skipping segment.",
                    array_element_read.query_name, si, ei, start_marker.end, end_marker.start)
            elif ei - si != 2:
                logger.warning(
                    "Read %s: start segment (i=%d) and end segment (i=%d) have more than one "
                    "segment between them.  Skipping segment.", array_element_read.query_name, si, ei)
            elif segments[si + 1].name != "random":
                logger.warning("Read %s: segment between start segment (i=%d) and end segment (i=%d) "
                               "is not a random segment.  Skipping segment.", array_element_read.query_name, si, ei)
            num_segments_skipped += 1

    return extracted_segment, num_segments_extracted, num_segments_skipped


def _extract_region_from_model_with_coding_region(array_element_read, segments, m, base_padding, extracted_bam_file, barcode_conf_file):
    # Get our marker segments:
    # NOTE: Even though this is a list, we only expect one segment to be here:
    extraction_segments = [s for s in segments if s.name == m.coding_region]

    if len(extraction_segments) == 0:
        logger.debug(f"Did not find coding region in read: %s: %s", array_element_read.query_name, segments)
        return False, 0, 1

    else:
        if len(extraction_segments) > 1:
            logger.debug(f"Found %d coding regions in read %s: %s.  Only looking at first coding region.  All Segments: %s",
                         len(extraction_segments), array_element_read.query_name, extraction_segments, segments)

        # Create an AlignedSegment to output:
        aligned_segment = _create_extracted_aligned_segment(
            array_element_read, extraction_segments[0], 0, base_padding, barcode_conf_file
        )

        if aligned_segment:
            extracted_bam_file.write(aligned_segment)
            return True, 1, 0
        else:
            return False, 0, 1


def _create_extracted_aligned_segment(read, seg_to_extract, start_offset, base_padding, barcode_conf_file):
    """Create a pysam.AlignedSegment object to store the information from the extracted bases."""

    start_coord = seg_to_extract.start + start_offset - base_padding
    end_coord = seg_to_extract.end + base_padding

    # Bounds check our coords:
    if start_coord < 0:
        logger.debug("Calculated start for %s would start before read begins.  Setting to 0.", read.query_name)
        start_coord = 0
    elif start_coord >= len(read.query_sequence):
        logger.warning("Start coord for %s would start after read.  Cannot process.", read.query_name)
        return None

    if end_coord < 0:
        logger.warning("End coord for %s would start before read.  Cannot process.", read.query_name)
        return None
    elif end_coord >= len(read.query_sequence):
        logger.debug("Calculated end for %s would start after read ends.  Setting to 0.", read.query_name)
        end_coord = len(read.query_sequence)-1

    if end_coord <= start_coord:
        logger.warning("Start coord for %s would start at or after end coord.  Cannot process.", read.query_name)
        return None

    # Create our segment:
    a = pysam.AlignedSegment()
    a.query_name = read.query_name
    # Add one to end_coord because coordinates are inclusive:
    a.query_sequence = f"{read.query_sequence[start_coord:end_coord+1]}"
    a.query_qualities = read.query_alignment_qualities[start_coord: end_coord + 1]
    a.tags = read.get_tags()
    a.flag = 4  # unmapped flag
    a.mapping_quality = 255
    a.set_tag(longbow.utils.constants.READ_ALTERED_NAME_TAG, f"{read.query_name}/{start_coord}_{end_coord}")

    # Write our barcode confidence to the file if we have to:
    if barcode_conf_file is not None and a.has_tag(longbow.utils.constants.READ_BARCODE_CONF_FACTOR_TAG):
        barcode_conf_file.write(f"{a.get_tag(longbow.utils.constants.READ_RAW_BARCODE_TAG)}\t{a.get_tag(longbow.utils.constants.READ_BARCODE_CONF_FACTOR_TAG)}\n")

    return a



import logging
import time
import sys

import pickle

import click
import click_log

import multiprocessing as mp

from ..utils import model as LongbowModel
from ..utils.model import LibraryModel

from ..utils import cli_utils

logging.basicConfig(stream=sys.stderr)
logger = logging.getLogger("model")
click_log.basic_config(logger)


@click.command(name="model")
@click_log.simple_verbosity_option(logger)
@click.option(
    "-l",
    "--list-models",
    cls=cli_utils.MutuallyExclusiveOption,
    mutually_exclusive=["dump"],
    required=False,
    is_flag=True,
    default=False,
    help="List the names of all models supported natively by this version of Longbow."
)
@click.option(
    "-d",
    "--dump",
    cls=cli_utils.MutuallyExclusiveOption,
    mutually_exclusive=["list_models"],
    required=False,
    type=str,
    help="Dump the details of a given model.  "
         "This command creates a set of files corresponding to the given model "
         "including: json representation, dot file representations, transmission matrix, state emission json file."
)
def main(list_models, dump):
    """Get information about built-in Longbow models."""

    logger.info("Invoked via: longbow %s", " ".join(sys.argv[1:]))

    if list_models or (not list_models and dump is None):
        print("Longbow includes the following models:")
        model_info_list = []
        col_widths = [0, 0, 0]
        for k, v in LibraryModel.pre_configured_models.items():
            model_info_list.append((k, v["version"], v["description"]))

            # Make sure we can print things nice and tidy:
            if len(k) > col_widths[0]:
                col_widths[0] = len(k)
            if len(v["version"]) > col_widths[1]:
                col_widths[1] = len(v["version"])

        model_name = "Name"
        version = "Version"
        description = "Description"
        print(f"{model_name:{col_widths[0] + 4}s}{version:{col_widths[1] + 4}s}{description}")
        for model_name, version, description in model_info_list:
            print(f"{model_name:{col_widths[0]+4}s}{version:{col_widths[1]+4}s}{description}")

    if dump is not None:
        model_name = dump
        if model_name not in LibraryModel.pre_configured_models.keys():
            logger.error(f"Given model name not in preconfigured models: {model_name} - {list(LibraryModel.pre_configured_models.keys())}")
        else:
            # Get our model:
            if LibraryModel.has_prebuilt_model(model_name):
                m = LibraryModel.build_pre_configured_model(model_name)
            else:
                logger.info(f"Loading model from json file: %s", model_name)
                m = LibraryModel.from_json_file(model_name)
            logger.info(f"Dumping %s: %s", model_name, m.description)

            model_dump_base_name = f"longbow_model_{m.name}.v{m.version}"

            logger.info(f"Dumping dotfile: {model_dump_base_name + '.dot'}")
            m.dump_as_dotfile(out_file_name=model_dump_base_name + ".dot", do_subgraphs=False)

            logger.info(f"Dumping simple dotfile: {model_dump_base_name + '.simple.dot'}")
            m.dump_as_dotfile_simple(out_file_name=model_dump_base_name + ".simple.dot")

            logger.info(f"Dumping json model specification: {model_dump_base_name + '.spec.json'}")
            m.to_json(f"{model_dump_base_name}.spec.json")

            logger.info(f"Dumping dense transition matrix: {model_dump_base_name + '.dense_transition_matrix.pickle'}")
            with open(f"{model_dump_base_name}.dense_transition_matrix.pickle", 'wb') as f:
                pickle.dump(m.hmm.dense_transition_matrix(), f)

            logger.info(f"Dumping emission distributions: {model_dump_base_name + '.emission_distributions.txt'}")
            with open(f"{model_dump_base_name}.emission_distributions.txt", 'w') as f:
                print(m.hmm, file=f, flush=True)

